#include "stdafx.h"

#include "OsiClp/OsiClpSolverInterface.hpp"
#include "CbcModel.hpp"
#include "CoinModel.hpp"

#include <cstdio>



int main(void)
{
	OsiClpSolverInterface solver;
	CbcModel model(solver);

	model.solver()->readLp("model.lp");

	// callCBC arguments
	std::string args("-clique off -knapsack off -mixed on -flow off -solve");

	// switch between fullsolve <--> iterative solve
	bool fullSolve = false;

	if(fullSolve) 
	{
		// full solve to show what's the real optimum = 381.86216 (N.B. it takes ~400 seconds to prove the optimality)  
		callCbc(args, model);
		if(model.bestSolution() != NULL && !model.isProvenInfeasible())
			printf(">>>>>>> BEST SOLUTION FOUND: %f\n", model.getObjValue());
		else
			printf(">>>>>>> IMPOSSIBLE TO FIND SOLUTION\n");
	}
	else  
	{
		// solve the problem by fixing variable iteratively and finally run a global solve
		// (N.B. the final solution I found with this approach is 383.118855, clearly not the global optimum)


		// columns to work on
		int columns2workon[] = {3,4,5,6,7,10,11,13,14,17,18,19,20,21,24,25,26,27,28};
		int numColumns2workon = 19;

		int numCols = model.getNumCols();

		// save original bounds and limits
		const double* saveLB = CoinCopyOfArray(model.getColLower(), numCols);
		const double* saveUB = CoinCopyOfArray(model.getColUpper(), numCols);

		int maxSol = model.getMaximumSolutions();
		int maxNodes = model.getMaximumNodes();

		double bestObj = COIN_DBL_MAX;
		double* bestSol = new double[numCols];

		// ****************************************************************************************************** //
		// find an initial feasible solution 
		// by forcing all columns2workon to 1, and stops at the first solution
		for(int i = 0; i < numColumns2workon; i++)
			model.solver()->setColBounds(columns2workon[i],1.0,1.0);
		model.setMaximumSolutions(1);

		callCbc(args, model);

		if(model.bestSolution() != NULL && !model.isProvenInfeasible())
		{
			bestObj = model.getObjValue();
			memcpy(bestSol, model.bestSolution(), numCols*sizeof(double));
			printf(">>>>>>> FIRST SOLUTION FOUND: %f\n", bestObj);
		}
		else
		{
			printf(">>>>>>> IMPOSSIBLE TO FIND INITIAL SOLUTION\n");
			exit(-1); // first solution has not been found
		}
		// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ //


		// ****************************************************************************************************** //
		// iterative improvement
		int improvements;
		do
		{
			improvements = 0;

			for(int i = 0; i < numColumns2workon; i++)
			{
				int colIndex2zero = columns2workon[i];

				// restore bounds and limits
				model.solver()->setColLower(saveLB);
				model.solver()->setColUpper(saveUB);
				model.setMaximumNodes(maxNodes);
				model.setMaximumSolutions(maxSol);

				// change col bounds by forcing them to the current best solution
				// except for colIndex, that is forced to 0
				for(int j = 0; j < numColumns2workon; j++)
				{
					int currColIndex = columns2workon[j];
					double val = bestSol[currColIndex];
					if(colIndex2zero == currColIndex)
						val = 0.0;

					model.solver()->setColBounds(currColIndex,val,val);
				}

				// solve for max 100 nodes
				model.setMaximumNodes(100);
				callCbc(args,model);

				if(model.bestSolution() != NULL && !model.isProvenInfeasible())
				{
					double currObj = model.getObjValue();
					if(currObj + 0.0000001 < bestObj)
					{
						bestObj = currObj;
						memcpy(bestSol, model.bestSolution(), numCols*sizeof(double));
						improvements++;

						printf(">>>>>>> IMPROVEMENT FOUND: %f\n", bestObj);
					}
				}
			}
		}
		while(improvements > 0);
		// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ //


		// ****************************************************************************************************** //
		// final solve

		// restore bounds and limits
		model.solver()->setColLower(saveLB);
		model.solver()->setColUpper(saveUB);
		model.setMaximumNodes(maxNodes);
		model.setMaximumSolutions(maxSol);

		callCbc(args,model);

		if(model.bestSolution() != NULL && !model.isProvenInfeasible())
		{
			bestObj = model.getObjValue();
			memcpy(bestSol, model.bestSolution(), numCols*sizeof(double));

			printf(">>>>>>> FINAL SOLUTION: %f\n", bestObj);
		}
		else
		{
			printf(">>>>>>> IMPOSSIBLE TO FIND FINAL SOLUTION\n");
		}

		// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ //

		delete[] bestSol;

	}

	printf("PRESS ANY KEY\n");
	getchar(); 

	return 0;
}

